from flask_classy import FlaskView
from flask import request, jsonify
from validators import TYPE_VALIDATOR
from marshmallow import *
from pagination import _validate_pagination_parameters, paginated_response
from sorting import _validate_sorting_parameters, sort
from filtering import _filter_query
from validators import _validate_params, REQUEST_BODY
from werkzeug.exceptions import BadRequest, NotFound

CREATE_NEW_OBJECT = "CREATE_NEW"
ASSOCIATE_EXISTING_OBJECT = "ASSOCIATE_EXISTING"


class FlaskImprovedView(FlaskView):
    route_base = "/"
    id_name = "id"
    view_schema = Schema()
    active_field = "active"
    view_model = None
    db_engine = None
    body_validation = {
        "sample_attribute": {
            "required": True,
            "validation_tuple": [(TYPE_VALIDATOR, unicode)]
        },
        "derivated_attribute": {
            "derivated": True,
            "object_type": None,
            "id_name": "id_derivated",
            "create_behavior": CREATE_NEW_OBJECT,
            "many": False,
            "fields": {
                "attribute1": {
                    "required": True,
                    "validation_tuple": [(TYPE_VALIDATOR, unicode)]
                }
            }
        }
    }

    def __process_derivated_attribute(self, derivated_data, derivated_validation, attr_name=None):
        object_type = derivated_validation.get("object_type", None)
        id_name = derivated_validation.get("id_name", None)
        create_behavior = derivated_validation.get("create_behavior", None)
        if derivated_validation.get("many", True) and type(derivated_data) == list:
            new_list_objects = list()
            for items_objects in derivated_data:
                new_object = self.__process_derivated_attribute(items_objects, derivated_validation)
                new_list_objects.append(new_object)
            return new_list_objects
        else:
            if create_behavior == CREATE_NEW_OBJECT:
                new_object = object_type()
                for key, value in derivated_validation.get("fields", {}).iteritems():
                    if value.get("derivated", False):
                        setattr(new_object, key, self.__process_derivated_attribute(derivated_data.get(key),
                                                                                    derivated_validation.get(key)))
                    else:
                        setattr(new_object, key, derivated_data.get(key))
                return new_object
            elif create_behavior == ASSOCIATE_EXISTING_OBJECT:
                existing_object = object_type.query.filter(getattr(object_type, id_name) == derivated_data.get(id_name, None)).first()
                if existing_object is None:
                    raise NotFound("Object of type %s with id %s not found in server. Try changing the id value."
                                   % (object_type, derivated_data.get(id_name, None)))
                return existing_object
        return None

    def index(self):
        _validate_sorting_parameters(request.args, self.view_model)
        _validate_pagination_parameters(request.args)
        result = self.view_model.query
        result = _filter_query(self.view_model, result, request.args.get("filter", None))
        sorted_result = sort(request.args, result)
        return paginated_response(request.args, sorted_result, type(self.view_schema))

    def get(self, id):
        result = self.view_model.query.get_or_404(int(id))
        return jsonify(self.view_schema.dump(result).data)

    def post(self):
        _validate_params(REQUEST_BODY, self.body_validation)
        data = request.json
        new_object = self.view_model()
        for key, value in data.iteritems():
            if type(value) in [list, dict]:
                setattr(new_object, key, self.__process_derivated_attribute(value, self.body_validation.get(key)))
            elif key != self.id_name and key in self.view_model.__table__.columns.keys():
                setattr(new_object, key, value)
        self.db_engine.session.add(new_object)
        try:
            self.db_engine.session.commit()
        except:
            self.db_engine.session.rollback()
            raise
        return jsonify(self.view_schema.dump(new_object).data), 201

    def put(self, id):
        _validate_params(REQUEST_BODY, self.body_validation)
        merged_object = self.view_model.query.get_or_404(int(id))
        data = request.json
        for key, value in data.iteritems():
            if type(value) in [list, dict]:
                setattr(merged_object, key, self.__process_derivated_attribute(value, self.body_validation.get(key)))
            elif key != self.id_name and key in self.view_model.__table__.columns.keys():
                setattr(merged_object, key, value)
        self.db_engine.session.merge(merged_object)
        try:
            self.db_engine.session.commit()
        except:
            self.db_engine.session.rollback()
            raise
        return jsonify(self.view_schema.dump(merged_object).data), 200

    def delete(self, id):
        deleted_object = self.view_model.query.get_or_404(int(id))
        if self.active_field is None or self.active_field not in self.view_model.__table__.columns.keys():
            self.db_engine.session.delete(deleted_object)
        else:
            setattr(deleted_object, self.active_field, False)
            self.db_engine.session.merge(deleted_object)
        try:
            self.db_engine.session.commit()
        except:
            self.db_engine.session.rollback()
            raise
        return jsonify({"message": "Successfully deleted item"}), 204




from werkzeug.exceptions import BadRequest, InternalServerError
from functools import update_wrapper
from flask import request
import re

TYPE_VALIDATOR = 1
MIN_VALIDATOR = 2
MAX_VALIDATOR = 3
REGEX_VALIDATOR = 4
VALID_VALUES_VALIDATOR = 5
NUMERIC_STRING_VALIDATOR = 6

REQUEST_QUERY_PARAMS = "ARGS"
REQUEST_BODY = "BODY"


def __validate_type(param_key, param_value, type):
    if isinstance(param_value, type):
        return param_value
    else:
        raise BadRequest("The value %s for param %s is not of type %s" % (str(param_value), param_key, type))


def __validate_min(param_key, param_value, min):
    if param_value >= min:
        return param_value
    else:
        raise BadRequest("The value %s for param %s is less than %s" % (str(param_value), param_key, min))


def __validate_max(param_key, param_value, max):
    if param_value <= max:
        return param_value
    else:
        raise BadRequest("The value %s for param %s is greater than %s" % (str(param_value), param_key, max))


def __validate_regex(param_key, param_value, regex):
    if re.match(regex, param_value):
        return param_value
    else:
        raise BadRequest("The value %s for param %s does not match to regex %s" % (str(param_value), param_key, regex))


def __validate_valid_values(param_key, param_value, valid_values):
    if param_value in valid_values:
        return param_value
    else:
        raise BadRequest("The value %s for param %s is not a valid value. Valid values: %s"
                         % (param_value, param_key, ', '.join(str(valid) for valid in valid_values)))


def __validate_numeric_string(param_key, param_value):
    try:
        float(param_value)
        return param_value
    except:
        raise BadRequest("The value %s for param %s is not a valid numeric value" % (param_value, param_key))


def validate_param_internal(data, param_key, param_tuple, required=False):
    param = data.get(param_key, None)
    if param is None:
        if not required:
            return None
        else:
            raise BadRequest("The param %s is mandatory" % param_key)

    for ituple in param_tuple:
        if ituple[0] == TYPE_VALIDATOR:
            param = __validate_type(param_key, param, ituple[1])
        elif ituple[0] == MIN_VALIDATOR:
            param = __validate_min(param_key, param, ituple[1])
        elif ituple[0] == MAX_VALIDATOR:
            param = __validate_max(param_key, param, ituple[1])
        elif ituple[0] == REGEX_VALIDATOR:
            param = __validate_regex(param_key, param, ituple[1])
        elif ituple[0] == VALID_VALUES_VALIDATOR:
            param = __validate_valid_values(param_key, param, ituple[1])
        elif ituple[0] == NUMERIC_STRING_VALIDATOR:
            param = __validate_numeric_string(param_key, param)


def validate_param(origin, param_key, param_tuple, required=False):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if origin == REQUEST_QUERY_PARAMS:
                data = request.args
            elif origin == REQUEST_BODY:
                data = request.json
            else:
                raise InternalServerError("An internal server error occurred during param %s validation." % param_key)
            validate_param_internal(data, param_key, param_tuple, required)
            return func(*args, **kwargs)
        return update_wrapper(wrapper, func)
    return decorator


def _validate_params(origin, params_dict, data=None):
    if data is None:
        if origin == REQUEST_QUERY_PARAMS:
            data = request.args
        elif origin == REQUEST_BODY:
            data = request.json
        else:
            raise InternalServerError("An internal server error occurred during param validation.")
    for param_key, param_data in params_dict.iteritems():
        if param_data.get("derivated", False):
            if type(param_data) == dict:
                if param_data.get("many", False):
                    for i_param_data in data.get(param_key, []):
                        _validate_params(origin, params_dict=param_data.get("fields", {}), data=i_param_data)
                else:
                    _validate_params(origin, params_dict=param_data.get("fields", {}), data=data.get(param_key))
        else:
            required = param_data.get("required", False)
            param_tuple = param_data.get("validation_tuple", [])
            validate_param_internal(data, param_key, param_tuple, required)


from werkzeug.exceptions import Unauthorized, Forbidden, BadRequest
from exceptions import AttributeError
import operator


class Authenticator:
    USER_OBJECT = None
    USERNAME_ATTR = "username"
    PASSWORD_ATTR = "password"
    SUPER_ROLE_ATTR = "role.is_super_role"
    PERMISSIONS_ATTR = "role.permissions"
    KEY_PERMISSION_ATTR = "user_permission_arn"

    def __init__(self):
        pass

    def __check_user_permissions(self, username, password, permission):
        user = self.USER_OBJECT.query.filter(getattr(self.USER_OBJECT, self.USERNAME_ATTR) == username)\
        .filter(getattr(self.USER_OBJECT, self.PASSWORD_ATTR) == password).first()
        if user is None:
            raise Unauthorized("The username might not exist or password is incorrect")
        if self.SUPER_ROLE_ATTR is not None:
            try:
                is_superuser = operator.attrgetter(self.SUPER_ROLE_ATTR)(user)
                if is_superuser:
                    return user
            except AttributeError:
                raise BadRequest("Internal error defining %s attribute. Not found in object." % self.SUPER_ROLE_ATTR)
        if self.PERMISSIONS_ATTR is not None and permission is not None:
            permissions = []
            try:
                permissions = operator.attrgetter(self.PERMISSIONS_ATTR)(user)
            except AttributeError:
                raise BadRequest("Internal error defining %s attribute. Not found in object." % self.PERMISSIONS_ATTR)
            try:
                permission_keys = [getattr(i_permission, self.KEY_PERMISSION_ATTR) for i_permission in permissions]
                if permission in permission_keys:
                    return user
                else:
                    raise Forbidden("User %s does not have permissions to execute %s" % (username, permission))
            except AttributeError:
                raise BadRequest("Internal error defining %s attribute. Not found in object."
                                 % self.KEY_PERMISSION_ATTR)
        return user

    def basic_auth(self, authorization, permission):
        if authorization is None or authorization.username is None or authorization.password is None:
            raise Unauthorized("There are missing authorization parameters. Please send user and password "
                               "in Authorization header.")
        user = self.__check_user_permissions(authorization.username, authorization.password, permission)
        return user



from setuptools import setup

setup(
    author='DinoCloud',
    author_email='info@dinocloudconsulting.com',
    name='flask_dino_utils',
    description='Flask utils package to use it among with Flask-Classy and marshmallow',
    url='http://github.com/dinocloud/flask-dino-utils',
    version='0.1.11',
    license='MIT',
    packages=['flask_dino_utils'],
    install_requires=[
        'Flask==0.11.1',
        'Flask-Classy==0.6.8',
        'marshmallow',
        'Flask-SQLAlchemy==1.0',
        'SQLAlchemy==1.0.16',
        'Werkzeug==0.11.11'
    ]
)
